// -*- coding: iso-8859-1 -*-

/**
* API - SDK Openbus C++
* \file openbus/Connection.h
* 
*/

#ifndef TECGRAF_CONNECTION_H_
#define TECGRAF_CONNECTION_H_

#include <CORBA.h>
#include <memory>
#include <stdexcept>
#include <string>
#include <vector>
#include <openssl/evp.h>
#include <cstring>

#include <boost/function.hpp>

#ifndef OPENBUS_DOXYGEN
#define SECRET_SIZE 16
#endif

namespace openbus {
  class Connection;
  struct CallerChain;
}

#include "openbus/interceptors/ORBInitializer_impl.h"
#include "openbus/interceptors/ClientInterceptor_impl.h"
#include "openbus/interceptors/ServerInterceptor_impl.h"
#include "openbus/Connection_impl.h"
#include "openbus/util/OpenSSL.h"
#include "openbus/ConnectionManager.h"

inline bool operator==(openbus::idl_ac::LoginInfo const &lhs, openbus::idl_ac::LoginInfo const &rhs)
{
  return lhs.id.in() == rhs.id.in() || 
    (lhs.id.in() && rhs.id.in() && !std::strcmp(lhs.id.in(), rhs.id.in()));
}

inline bool operator!=(openbus::idl_ac::LoginInfo const &lhs, openbus::idl_ac::LoginInfo const &rhs)
{
  return !(lhs == rhs);
}

/**
* \brief openbus
*/
namespace openbus {
  
/* exceptions */
struct BusChanged : public std::exception { 
  const char *what() const throw() { return "openbus::BusChanged"; }
};

struct AlreadyLoggedIn : public std::exception { 
  const char *what() const throw() { return "openbus::AlreadyLoggedIn"; } 
};

struct InvalidBusAddress : public std::exception { 
  const char *what() const throw() { return "openbus::InvalidBusAddress"; }
};

struct InvalidPrivateKey : public std::exception { 
  const char *what() const throw() { return "openbus::InvalidPrivateKey";} 
};

struct InvalidLoginProcess : public std::exception { 
  const char *what() const throw() { return "openbus::InvalidLoginProcess"; }
};

struct InvalidPropertyValue : public std::exception {
  InvalidPropertyValue(std::string p, std::string v) : property(p), value(v) { }
  ~InvalidPropertyValue() throw() { }
  const char *what() const throw() { return "openbus::InvalidBusAddress"; }
  std::string property;
  std::string value;
};
/**/

class Connection;

/**
 * \brief Cadeia de chamadas oriundas de um barramento.
 * 
 * Coleo de informaes dos logins que originaram chamadas em cadeia atravs de um
 * barramento. Cadeias de chamadas representam chamadas aninhadas dentro do barramento e so teis
 * para que os sistemas que recebam essas chamadas possam identificar se a chamada foi originada por
 * entidades autorizadas ou no.
 */
struct CallerChain {
  /**
  * Barramento atravs do qual as chamadas foram originadas.
  */
  const char *busid() const { return _busid.c_str(); }
  
  /**
  * Lista de informaes de login de todas as entidades que realizaram chamadas que originaram a
  * cadeia de chamadas da qual essa chamada est inclusa.  Quando essa lista  vazia isso indica
  * que a chamada no est inclusa em uma cadeia de chamadas.
  * 
  * A ordem da sequncia retornada  comeando da fonte da cadeia at o penltimo da cadeia na
  * chamada. Assim, originators()[0], se existir,  quem originou a chamada de cadeia.
  */
  const idl_ac::LoginInfoSeq &originators() const { return _originators; }
  
  /**
   * Informao de login da entidade que realizou a ltima chamada da cadeia.
   */
  const idl_ac::LoginInfo &caller() const { return _caller; }

  /**
   * \brief Construtor default que indica o valor de CallChain "vazio"
   *
   * O valor de um CallerChain default-constructed pode ser usado para verificar a ausencia de
   * CallerChain da seguinte forma:
   * 
   * CallerChain chain = connection.getCallerChain();
   * if(chain != CallerChain())
   *   // Possui CallerChain
   * else
   *   // Nao possui CallerChain
   *
   */
  CallerChain() {}
private:
  CallerChain(const char *busid, const idl_ac::LoginInfoSeq &b, const idl_ac::LoginInfo &c, 
    const idl_cr::SignedCallChain &d) 
    : _busid(busid), _originators(b), _caller(c), _signedCallChain(d) { }
  
  CallerChain(const char *busid, const idl_ac::LoginInfoSeq &b, const idl_ac::LoginInfo &c) 
    : _busid(busid), _originators(b), _caller(c) { }
  
  std::string _busid;
  idl_ac::LoginInfoSeq _originators;
  idl_ac::LoginInfo _caller;
  idl_cr::SignedCallChain _signedCallChain;
  const idl_cr::SignedCallChain *signedCallChain() const { return &_signedCallChain; }
  void signedCallChain(idl_cr::SignedCallChain p) { _signedCallChain = p; }
  friend class Connection;
  friend class openbus::interceptors::ClientInterceptor;
  friend inline bool operator==(CallerChain const &lhs, CallerChain const &rhs) {
    return lhs._busid == rhs._busid && lhs._originators == rhs._originators
      && lhs._caller == rhs._caller;
  }
};

inline bool operator!=(CallerChain const &lhs, CallerChain const &rhs) { return !(lhs == rhs); }

/**
 * \brief Objeto que representa uma forma de acesso a um barramento.
 *
 * Uma conexo representa uma forma de acesso a um barramento. Basicamente, uma conexo  usada para
 * representar uma identidade de acesso a um barramento.   possvel uma aplicao assumir mltiplas
 * identidades ao acessar um ou mais barramentos criando mltiplas conexes para esses barramentos.
 * 
 * Para que as conexes possam ser efetivamente utilizadas elas precisam estar autenticadas no
 * barramento, que pode ser visto como um identificador de acesso. Cada login possui um
 * identificador nico e  autenticado em nome de uma entidade, que pode representar um sistema
 * computacional ou mesmo uma pessoa. A funo da entidade  atribuir a responsabilidade as chamadas
 * feitas com aquele login.
 * 
 *  importante notar que a conexo define uma forma de acesso, mas no  usada diretamente pela
 * aplicao ao realizar ou receber chamadas, pois as chamadas ocorrem usando proxies e servants de
 * um ORB. As conexes que so efetivamente usadas nas chamadas do ORB so definidas atravs do
 * ConnectionManager associado ao ORB.
 */
class Connection {
public:
  /**
   * \brief Callback de login invlido.
   * 
   * Tipo que representa um objeto funo ('function object') a ser chamado quando uma notificao
   * de login invlido  recebida. Caso alguma exceo ocorra durante a execuo do mtodo e no
   * seja tratada, o erro ser capturado pelo interceptador e registrado no log.
   * 
   * O tipo InvalidLoginCallback_t  um typedef de boost::function. Para documentao dessa
   * biblioteca acesse http://www.boost.org/doc/libs/1_47_0/doc/html/function.html
   *
   * \param conn Conexo que recebeu a notificao de login invlido.
   * \param login Informaes do login que se tornou invlido.
   */
  typedef boost::function<void (Connection&, idl_ac::LoginInfo)> InvalidLoginCallback_t;
  
  /**
  * Efetua login no barramento como uma entidade usando autenticao por senha.
  * 
  * @param[in] entity Identificador da entidade a ser autenticada.
  * @param[in] password Senha de autenticao no barramento da entidade.
  * 
  * @throw AlreadyLoggedIn A conexo j est logada.
  * @throw BusChanged O identificador do barramento mudou. Uma nova conexo deve ser criada.
  * @throw tecgraf::openbus::core::v2_0::services::access_control::AccessDenied
  *        Senha fornecida para autenticao da entidade no foi validada pelo barramento.
  * @throw tecgraf::openbus::core::v2_0::services::ServiceFailure 
  *        Ocorreu uma falha interna nos servios do barramento que impediu a autenticao da 
  *        conexo.
  * @throw CORBA::Exception
  */
  void loginByPassword(const char *entity, const char *password);
  
 /**
  * \brief Efetua login de uma entidade usando autenticao por certificado.
  * 
  * A autenticao por certificado  validada usando um certificado de login registrado pelo
  * adminsitrador do barramento.
  * 
  * @param[in] entity Identificador da entidade a ser conectada.
  * @param[in] privKey Chave privada da entidade utilizada na autenticao.
  * 
  * @throw InvalidPrivateKey A chave privada fornecida no  vlida.
  * @throw AlreadyLoggedIn A conexo j est autenticada.
  * @throw BusChanged O identificador do barramento mudou. Uma nova conexo deve ser criada.
  * @throw tecgraf::openbus::core::v2_0::services::access_control::AccessDenied 
  *        A chave privada fornecida no corresponde ao certificado da entidade registrado no
  *        barramento indicado.
  * @throw MissingCertificate No h certificado para essa entidade registrado no barramento 
  *        indicado.
  * @throw tecgraf::openbus::core::v2_0::services::ServiceFailure 
  *        Ocorreu uma falha interna nos servios do barramento que impediu a autenticao da 
  *        conexo.
  * @throw CORBA::Exception
  */
  void loginByCertificate(const char *entity, const idl::OctetSeq &privKey);
  
  /**
  * \brief Inicia o processo de login por autenticao compartilhada.
  * 
  * A autenticao compartilhada permite criar um novo login compartilhando a mesma autenticao do
  * login atual da conexo. Portanto essa operao s pode ser chamada enquanto a conexo estiver
  * autenticada, caso contrrio a exceo de sistema CORBA::NO_PERMISSION{NoLogin}  lanada. As
  * informaes fornecidas por essa operao devem ser passadas para a operao 'loginBySharedAuth'
  * para concluso do processo de login por autenticao compartilhada. Isso deve ser feito dentro
  * do tempo de lease definido pelo administrador do barramento. Caso contrrio essas informaes se
  * tornam invlidas e no podem mais ser utilizadas para criar um login.
  * 
  * @return Um par composto de um objeto que representa o processo de login iniciado e de um segredo
  *         a ser fornecido na concluso do processo de login.
  *
  * @throw tecgraf::openbus::core::v2_0::services::ServiceFailure 
  *        Ocorreu uma falha interna nos servios do barramento que impediu o estabelecimento da
  *        conexo.
  * @throw CORBA::Exception
  */
  std::pair <idl_ac::LoginProcess_ptr, idl::OctetSeq> startSharedAuth();
  
  /**
  * \brief Efetua login de uma entidade usando autenticao compartilhada.
  * 
  * A autenticao compartilhada  feita a partir de informaes obtidas a atravs da operao
  * 'startSharedAuth' de uma conexo autenticada.
  * 
  * @param[in] loginProcess Objeto que represeta o processo de login iniciado.
  * @param[in] secret Segredo a ser fornecido na concluso do processo de login.
  * 
  * @throw InvalidLoginProcess O LoginProcess informado  invlido, por exemplo depois de ser 
  *        cancelado ou ter expirado.
  * @throw AlreadyLoggedIn A conexo j est autenticada.
  * @throw BusChangedO identificador do barramento mudou. Uma nova conexo deve ser criada.
  * @throw tecgraf::openbus::core::v2_0::services::access_control::AccessDenied 
  *        O segredo fornecido no corresponde ao esperado pelo barramento.
  * @throw tecgraf::openbus::core::v2_0::services::ServiceFailure 
  *        Ocorreu uma falha interna nos serviços do barramento que impediu o estabelecimento 
  *        da conexo.
  * @throw CORBA::Exception
  */
  void loginBySharedAuth(idl_ac::LoginProcess_ptr loginProcess, const idl::OctetSeq &secret);
  
 /**
  * \brief Efetua logout da conexo, tornando o login atual invlido.
  * 
  * Aps a chamada a essa operao a conexo fica desautenticada, implicando que qualquer chamada
  * realizada pelo ORB usando essa conexo resultar numa exceo de sistema
  * 'CORBA::NO_PERMISSION{NoLogin}' e chamadas recebidas por esse ORB sero respondidas com a
  * exceo 'CORBA::NO_PERMISSION{UnknownBus}' indicando que no foi possvel validar a chamada pois
  * a conexo est temporariamente desautenticada.
  * 
  * @return Verdadeiro se o processo de logout for concludo com xito e falso se a conexo j 
  *         estiver desautenticada (login invlido).
  *
  * @throw CORBA::Exception
  */
  bool logout();
  
  /**
   * \brief Devolve a cadeia de chamadas  qual a execuo corrente pertence.
   * 
   * Caso a contexto corrente (e.g. definido pelo 'CORBA::PICurrent') seja o contexto de execuo de
   * uma chamada remota oriunda do barramento dessa conexo, essa operao devolve um objeto que
   * representa a cadeia de chamadas do barramento que esta chamada faz parte. Caso contrrio,
   * devolve uma cadeia de chamadas 'vazia', representada por um CallerChain 'default-constructed'.
   *
   * Para verificar se a cadeia retornada  vlida, o seguinte idioma
   *  usado:
   *
   * CallerChain chain = connection.getCallerChain()
   * if(chain != CallerChain())
   *   // chain  vlido
   * else
   *   // chain  invlido
   * 
   * \return Cadeia da chamada em execuo.
   */
  CallerChain getCallerChain();
  
  /**
   * \brief Associa uma cadeia de chamadas ao contexto corrente.
   * 
   * Associa uma cadeia de chamadas ao contexto corrente, de forma que todas as chamadas remotas
   * seguintes neste mesmo contexto sejam feitas como parte dessa cadeia de chamadas.
   * 
   * \param chain Cadeia de chamadas a ser associada ao contexto corrente.
   * @throw CORBA::NO_PERMISSION {minor = NoLoginCode}
   * @throw CORBA::NO_PERMISSION {minor = InvalidChainCode}
   * @throw CORBA::Exception
   */
  void joinChain(CallerChain const& chain);
  
  /**
   * \brief Faz com que nenhuma cadeia de chamadas esteja associada ao contexto corrente.
   * 
   * Remove a associao da cadeia de chamadas ao contexto corrente, fazendo com que todas as
   * chamadas seguintes feitas neste mesmo contexto deixem de fazer parte da cadeia de chamadas
   * associada previamente. Ou seja, todas as chamadas passam a iniciar novas cadeias de chamada.
   */
  void exitChain();

  /**
   * \brief Devolve a cadeia de chamadas associada ao contexto corrente.
   * 
   * Devolve um objeto que representa a cadeia de chamadas associada ao contexto corrente nesta
   * conexo.  A cadeia de chamadas informada foi associada previamente pela operao
   * 'joinChain'. Caso o contexto corrente no tenha nenhuma cadeia associada, essa operao devolve
   * uma cadeia 'vazia' 'default-constructed'
   * 
   * Para verificar se a cadeia retornada  vlida, o seguinte idioma  usado:
   *
   * CallerChain chain = connection.getCallerChain()
   * if(chain != CallerChain())
   *   // chain  vlido
   * else
   *   // chain  invlido
   * 
   * \return Cadeia de chamadas associada ao contexto corrente ou uma cadeia 'vazia'.
   *
   * @throw CORBA::Exception
   */
  CallerChain getJoinedChain();
  
	
  /**
   * \brief Estabiliza a callback a ser chamada quando o login atual se tornar invlido.
   *
   * Esse atributo  utilizado para definir um objeto funo (function object) que implementa uma
   * interface de callback a ser chamada sempre que a conexo receber uma notificao de que o seu
   * login est invlido. Essas notificaes ocorrem durante chamadas realizadas ou recebidas pelo
   * barramento usando essa conexo. Um login pode se tornar invlido caso o administrador
   * explicitamente o torne invlido ou caso a thread interna de renovao de login no seja capaz
   * de renovar o lease do login a tempo. Caso esse atributo seja um InvalidLoginCallback_t
   * 'default-constructed', nenhum objeto de callback  chamado na ocorrncia desse evento.
   *
   * Durante a execuo dessa callback um novo login pode ser restabelecido.  Neste caso, a chamada
   * do barramento que recebeu a notificao de login invlido  refeita usando o novo login, caso
   * contrrio, a chamada original lana a exceo de de sistema 'CORBA::NO_PERMISSION{NoLogin}'.
   * 
   * O tipo InvalidLoginCallback_t  um typedef de boost::function. Para documentao dessa
   * biblioteca acesse http://www.boost.org/doc/libs/1_47_0/doc/html/function.html
   */
  void onInvalidLogin(InvalidLoginCallback_t p);
  
  /**
   * \brief Retorna a callback configurada para ser chamada quando o login atual se torna invlido.
   */
  InvalidLoginCallback_t onInvalidLogin();
  
  /**
   * \brief Informaes do login dessa conexo ou 'null' se a conexo no est autenticada, ou seja,
   * no tem um login vlido no barramento.
   */
  const idl_ac::LoginInfo* login();
  
  /**
   * Identificador do barramento ao qual essa conexo se refere.
   */
  const char *busid();
  
  /**
   * Referncia ao servio ncleo de registro de ofertas do barramento ao qual a conexo se refere.
   */
  idl_or::OfferRegistry_var offers() const { return _offer_registry; }
  
  ~Connection();
private:
  /**
  * Connection deve ser adquirido atravs de: ConnectionManager::createConnection()
  */
  Connection(const std::string host, const unsigned short port, CORBA::ORB*, IOP::Codec*, 
    PortableInterceptor::SlotId slotId_joinedCallChain, 
    PortableInterceptor::SlotId slotId_signedCallChain, 
    PortableInterceptor::SlotId slotId_legacyCallChain, 
    PortableInterceptor::SlotId slotId_receiveConnection, 
    ConnectionManager*, std::vector<std::string> props);
  openssl::pkey fetchBusKey();
  void checkBusid() const;
  bool _logout(bool local);
  openssl::pkey __key() const { return _key; }
  openssl::pkey __buskey() const { return _buskey; }
  CORBA::ORB *orb() const { return _orb; }
  idl_ac::LoginRegistry_var login_registry() const { return _login_registry; }
  idl_ac::AccessControl_var access_control() const { return _access_control; }
  const idl_ac::LoginInfo *_login() const { return _loginInfo.get(); }
  const std::string _host;
  const unsigned short _port;
  CORBA::ORB *_orb;
  IOP::Codec *_codec;
  PortableInterceptor::SlotId _slotId_joinedCallChain; 
  PortableInterceptor::SlotId _slotId_signedCallChain;
  PortableInterceptor::SlotId _slotId_legacyCallChain;
  PortableInterceptor::SlotId _slotId_receiveConnection;
  std::auto_ptr<RenewLogin> _renewLogin;
  std::auto_ptr<idl_ac::LoginInfo> _loginInfo;
  InvalidLoginCallback_t _onInvalidLogin;
  Mutex _mutex;
  
  enum LegacyDelegate {
    CALLER,
    ORIGINATOR
  };
  
  enum State {
    LOGGED,
    UNLOGGED,
    INVALID
  } _state;
  
  /* Variveis que so modificadas somente no construtor. */
  ConnectionManager *_manager;
  openssl::pkey _key;
  PortableInterceptor::Current_var _piCurrent;
  scs::core::IComponent_var _iComponent;
  idl_ac::AccessControl_var _access_control;
  idl_ac::LoginRegistry_var _login_registry;
  idl_or::OfferRegistry_var _offer_registry;
  std::auto_ptr<LoginCache> _loginCache;
  const char *_busid;
  openssl::pkey _buskey;
  LegacyDelegate _legacyDelegate;
  /**/
  
  friend class openbus::interceptors::ServerInterceptor;
  friend class openbus::interceptors::ClientInterceptor;
  friend class openbus::ConnectionManager;
};

}

#endif
